@@ -13,18 +13,23 @@
 ##############################################################################
 
 import logging
-from cStringIO import StringIO
-from cPickle import Unpickler, Pickler
-from pickle import PicklingError
+import sys
 
+import six
 import zope.interface
-
 from ZODB.POSException import ConflictError
 from ZODB.loglevels import BLATHER
+from ZODB._compat import BytesIO, Unpickler, Pickler, _protocol
+
+# Subtle: Python 2.x has pickle.PicklingError and cPickle.PicklingError,
+# and these are unrelated classes!  So we shouldn't use pickle.PicklingError,
+# since on Python 2, ZODB._compat.pickle is cPickle.
+from pickle import PicklingError
+
 
 logger = logging.getLogger('ZODB.ConflictResolution')
 
-ResolvedSerial = 'rs'
+ResolvedSerial = b'rs'
 
 class BadClassName(Exception):
     pass
@@ -57,8 +62,8 @@ def find_global(*args):
     if cls == 1:
         # Not importable
         if (isinstance(args, tuple) and len(args) == 2 and
-            isinstance(args[0], basestring) and
-            isinstance(args[1], basestring)
+            isinstance(args[0], six.string_types) and
+            isinstance(args[1], six.string_types)
             ):
             return BadClass(*args)
         else:
@@ -68,7 +73,7 @@ def find_global(*args):
 def state(self, oid, serial, prfactory, p=''):
     p = p or self.loadSerial(oid, serial)
     p = self._crs_untransform_record_data(p)
-    file = StringIO(p)
+    file = BytesIO(p)
     unpickler = Unpickler(file)
     unpickler.find_global = find_global
     unpickler.persistent_load = prfactory.persistent_load
@@ -112,9 +117,9 @@ class IPersistentReference(zope.interface.Interface):
         have two references to the same object that are spelled with different
         data (for instance, one with a class and one without).'''
 
+@zope.interface.implementer(IPersistentReference)
 class PersistentReference(object):
 
-    zope.interface.implements(IPersistentReference)
 
     weak = False
     oid = database_name = klass = None
@@ -130,7 +135,7 @@ class PersistentReference(object):
                 # it.  Fortunately, a class reference in a persistent
                 # reference is allowed to be a module+name tuple.
                 self.data = self.oid, klass.args
-        elif isinstance(data, str):
+        elif isinstance(data, (bytes, str)):
             self.oid = data
         else: # a list
             reference_type = data[0]
@@ -140,7 +145,10 @@ class PersistentReference(object):
             #    or persistent weakref: (oid, database_name)
             # else it is a weakref: reference_type
             if reference_type == 'm':
-                self.database_name, self.oid, _ = data[1]
+                self.database_name, self.oid, klass = data[1]
+                if isinstance(klass, BadClass):
+                    # see above wrt BadClass
+                    data[1] = self.database_name, self.oid, klass.args
             elif reference_type == 'n':
                 self.database_name, self.oid = data[1]
             elif reference_type == 'w':
@@ -153,6 +161,10 @@ class PersistentReference(object):
                 assert len(data) == 1, 'unknown reference format'
                 self.oid = data[0]
                 self.weak = True
+        if not isinstance(self.oid, (bytes, type(None))):
+            assert isinstance(self.oid, str)
+            # this happens on Python 3 when all bytes in the oid are < 0x80
+            self.oid = self.oid.encode('ascii')
 
     def __cmp__(self, other):
         if self is other or (
@@ -167,13 +179,32 @@ class PersistentReference(object):
                 "can't reliably compare against different "
                 "PersistentReferences")
 
+    # Python 3 dropped __cmp__
+
+    def __eq__(self, other):
+        return self.__cmp__(other) == 0
+
+    def __ne__(self, other):
+        return self.__cmp__(other) != 0
+
+    def __lt__(self, other):
+        return self.__cmp__(other) < 0
+
+    def __gt__(self, other):
+        return self.__cmp__(other) > 0
+
+    def __le__(self, other):
+        return self.__cmp__(other) <= 0
+
+    def __ge__(self, other):
+        return self.__cmp__(other) >= 0
+
     def __repr__(self):
         return "PR(%s %s)" % (id(self), self.data)
 
     def __getstate__(self):
         raise PicklingError("Can't pickle PersistentReference")
 
-
     @property
     def klass(self):
         # for tests
@@ -206,12 +237,12 @@ def persistent_id(object):
 
 _unresolvable = {}
 def tryToResolveConflict(self, oid, committedSerial, oldSerial, newpickle,
-                         committedData=''):
+                         committedData=b''):
     # class_tuple, old, committed, newstate = ('',''), 0, 0, 0
     try:
         prfactory = PersistentReferenceFactory()
         newpickle = self._crs_untransform_record_data(newpickle)
-        file = StringIO(newpickle)
+        file = BytesIO(newpickle)
         unpickler = Unpickler(file)
         unpickler.find_global = find_global
         unpickler.persistent_load = prfactory.persistent_load
@@ -254,12 +285,15 @@ def tryToResolveConflict(self, oid, committedSerial, oldSerial, newpickle,
 
         resolved = resolve(old, committed, newstate)
 
-        file = StringIO()
-        pickler = Pickler(file,1)
-        pickler.inst_persistent_id = persistent_id
+        file = BytesIO()
+        pickler = Pickler(file, _protocol)
+        if sys.version_info[0] < 3:
+            pickler.inst_persistent_id = persistent_id
+        else:
+            pickler.persistent_id = persistent_id
         pickler.dump(meta)
         pickler.dump(resolved)
-        return self._crs_transform_record_data(file.getvalue(1))
+        return self._crs_transform_record_data(file.getvalue())
     except (ConflictError, BadClassName):
         pass
     except:
@@ -268,7 +302,7 @@ def tryToResolveConflict(self, oid, committedSerial, oldSerial, newpickle,
         # the original ConflictError.  A client can recover from a
         # ConflictError, but not necessarily from other errors.  But log
         # the error so that any problems can be fixed.
-        logger.error("Unexpected error", exc_info=True)
+        logger.exception("Unexpected error")
 
     raise ConflictError(oid=oid, serials=(committedSerial, oldSerial),
                         data=newpickle)
