@@ -302,8 +302,7 @@ class Connection(ExportImport, object):
                     f()
                 except: # except what?
                     f = getattr(f, 'im_self', f)
-                    self._log.error("Close callback failed for %s", f,
-                                    exc_info=sys.exc_info())
+                    self._log.exception("Close callback failed for %s", f)
             self.__onCloseCallbacks = None
 
         self._debug_info = ()
@@ -874,8 +873,7 @@ class Connection(ExportImport, object):
             raise
         except:
             self._log.exception("Couldn't load state for %s %s",
-                                className(obj), oid_repr(oid),
-                                exc_info=sys.exc_info())
+                                className(obj), oid_repr(oid))
             raise
 
     def _setstate(self, obj):
