@@ -13,40 +13,78 @@
 ##############################################################################
 """Storage implementation using a log written to a single file.
 """
+from __future__ import print_function
 
-from __future__ import with_statement
-
-from cPickle import Pickler, loads
-from persistent.TimeStamp import TimeStamp
-from struct import pack, unpack
-from zc.lockfile import LockFile
-from ZODB.FileStorage.format import CorruptedError, CorruptedDataError
-from ZODB.FileStorage.format import FileStorageFormatter, DataHeader
-from ZODB.FileStorage.format import TRANS_HDR, TRANS_HDR_LEN
-from ZODB.FileStorage.format import TxnHeader, DATA_HDR, DATA_HDR_LEN
-from ZODB.FileStorage.fspack import FileStoragePacker
-from ZODB.fsIndex import fsIndex
-from ZODB import BaseStorage, ConflictResolution, POSException
-from ZODB.POSException import UndoError, POSKeyError, MultipleUndoErrors
-from ZODB.utils import p64, u64, z64
-
-import base64
-import BTrees.OOBTree
+import binascii
 import contextlib
 import errno
 import logging
 import os
 import threading
 import time
-import ZODB.blob
-import ZODB.interfaces
-import zope.interface
-import ZODB.utils
+from struct import pack
+from struct import unpack
+
+from persistent.TimeStamp import TimeStamp
+from six import string_types as STRING_TYPES
+from zc.lockfile import LockFile
+from zope.interface import alsoProvides
+from zope.interface import implementer
+
+from ZODB.blob import BlobStorageMixin
+from ZODB.blob import link_or_copy
+from ZODB.blob import remove_committed
+from ZODB.blob import remove_committed_dir
+from ZODB.BaseStorage import BaseStorage
+from ZODB.BaseStorage import DataRecord as _DataRecord
+from ZODB.BaseStorage import TransactionRecord as _TransactionRecord
+from ZODB.ConflictResolution import ConflictResolvingStorage
+from ZODB.ConflictResolution import ResolvedSerial
+from ZODB.FileStorage.format import CorruptedDataError
+from ZODB.FileStorage.format import CorruptedError
+from ZODB.FileStorage.format import DATA_HDR
+from ZODB.FileStorage.format import DATA_HDR_LEN
+from ZODB.FileStorage.format import DataHeader
+from ZODB.FileStorage.format import FileStorageFormatter
+from ZODB.FileStorage.format import TRANS_HDR
+from ZODB.FileStorage.format import TRANS_HDR_LEN
+from ZODB.FileStorage.format import TxnHeader
+from ZODB.FileStorage.fspack import FileStoragePacker
+from ZODB.interfaces import IBlobStorageRestoreable
+from ZODB.interfaces import IExternalGC
+from ZODB.interfaces import IStorage
+from ZODB.interfaces import IStorageCurrentRecordIteration
+from ZODB.interfaces import IStorageIteration
+from ZODB.interfaces import IStorageRestoreable
+from ZODB.interfaces import IStorageUndoable
+from ZODB.POSException import ConflictError
+from ZODB.POSException import MultipleUndoErrors
+from ZODB.POSException import POSKeyError
+from ZODB.POSException import ReadOnlyError
+from ZODB.POSException import StorageError
+from ZODB.POSException import StorageSystemError
+from ZODB.POSException import StorageTransactionError
+from ZODB.POSException import UndoError
+from ZODB.fsIndex import fsIndex
+from ZODB.utils import as_bytes
+from ZODB.utils import as_text
+from ZODB.utils import cp
+from ZODB.utils import mktemp
+from ZODB.utils import p64
+from ZODB.utils import u64
+from ZODB.utils import z64
+from ZODB._compat import Pickler
+from ZODB._compat import loads
+from ZODB._compat import decodebytes
+from ZODB._compat import encodebytes
+from ZODB._compat import _protocol
+from ZODB._compat import FILESTORAGE_MAGIC
+
 
 # Not all platforms have fsync
 fsync = getattr(os, "fsync", None)
 
-packed_version = "FS21"
+packed_version = FILESTORAGE_MAGIC
 
 logger = logging.getLogger('ZODB.FileStorage')
 
@@ -54,7 +92,7 @@ def panic(message, *data):
     logger.critical(message, *data)
     raise CorruptedTransactionError(message % data)
 
-class FileStorageError(POSException.StorageError):
+class FileStorageError(StorageError):
     pass
 
 class PackError(FileStorageError):
@@ -67,14 +105,14 @@ class FileStorageFormatError(FileStorageError):
     """
 
 class CorruptedFileStorageError(FileStorageError,
-                                POSException.StorageSystemError):
+                                StorageSystemError):
     """Corrupted file storage."""
 
 class CorruptedTransactionError(CorruptedFileStorageError):
     pass
 
 class FileStorageQuotaError(FileStorageError,
-                            POSException.StorageSystemError):
+                            StorageSystemError):
     """File storage quota exceeded."""
 
 # Intended to be raised only in fspack.py, and ignored here.
@@ -87,21 +125,21 @@ class TempFormatter(FileStorageFormatter):
     def __init__(self, afile):
         self._file = afile
 
+@implementer(
+        IStorage,
+        IStorageRestoreable,
+        IStorageIteration,
+        IStorageUndoable,
+        IStorageCurrentRecordIteration,
+        IExternalGC,
+        )
 class FileStorage(
     FileStorageFormatter,
-    ZODB.blob.BlobStorageMixin,
-    ConflictResolution.ConflictResolvingStorage,
-    BaseStorage.BaseStorage,
+    BlobStorageMixin,
+    ConflictResolvingStorage,
+    BaseStorage,
     ):
 
-    zope.interface.implements(
-        ZODB.interfaces.IStorage,
-        ZODB.interfaces.IStorageRestoreable,
-        ZODB.interfaces.IStorageIteration,
-        ZODB.interfaces.IStorageUndoable,
-        ZODB.interfaces.IStorageCurrentRecordIteration,
-        ZODB.interfaces.IExternalGC,
-        )
 
     # Set True while a pack is in progress; undo is blocked for the duration.
     _pack_is_in_progress = False
@@ -118,7 +156,7 @@ class FileStorage(
             raise ValueError("time-travel only supported in read-only mode")
 
         if stop is None:
-            stop='\377'*8
+            stop = b'\377'*8
 
         # Lock the database and set up the temp file.
         if not read_only:
@@ -136,7 +174,7 @@ class FileStorage(
         if packer is not None:
             self.packer = packer
 
-        BaseStorage.BaseStorage.__init__(self, file_name)
+        BaseStorage.__init__(self, file_name)
 
         index, tindex = self._newIndexes()
         self._initIndex(index, tindex)
@@ -147,7 +185,7 @@ class FileStorage(
         if not create:
             try:
                 self._file = open(file_name, read_only and 'rb' or 'r+b')
-            except IOError, exc:
+            except IOError as exc:
                 if exc.errno == errno.EFBIG:
                     # The file is too big to open.  Fail visibly.
                     raise
@@ -209,20 +247,19 @@ class FileStorage(
         if blob_dir:
             self.blob_dir = os.path.abspath(blob_dir)
             if create and os.path.exists(self.blob_dir):
-                ZODB.blob.remove_committed_dir(self.blob_dir)
+                remove_committed_dir(self.blob_dir)
 
             self._blob_init(blob_dir)
-            zope.interface.alsoProvides(self,
-                                        ZODB.interfaces.IBlobStorageRestoreable)
+            alsoProvides(self, IBlobStorageRestoreable)
         else:
             self.blob_dir = None
             self._blob_init_no_blobs()
 
     def copyTransactionsFrom(self, other):
         if self.blob_dir:
-            return ZODB.blob.BlobStorageMixin.copyTransactionsFrom(self, other)
+            return BlobStorageMixin.copyTransactionsFrom(self, other)
         else:
-            return BaseStorage.BaseStorage.copyTransactionsFrom(self, other)
+            return BaseStorage.copyTransactionsFrom(self, other)
 
     def _initIndex(self, index, tindex):
         self._index=index
@@ -354,7 +391,7 @@ class FileStorage(
         pos = info.get('pos')
         if index is None or pos is None:
             return None
-        pos = long(pos)
+        pos = int(pos)
 
         if (isinstance(index, dict) or
                 (isinstance(index, fsIndex) and
@@ -368,7 +405,7 @@ class FileStorage(
             if not self._is_read_only:
                 # Save the converted index.
                 f = open(index_name, 'wb')
-                p = Pickler(f, 1)
+                p = Pickler(f, _protocol)
                 info['index'] = index
                 p.dump(info)
                 f.close()
@@ -392,7 +429,7 @@ class FileStorage(
             self._save_index()
         except:
             # Log the error and continue
-            logger.error("Error saving index on close()", exc_info=True)
+            logger.exception("Error saving index on close()")
 
     def getSize(self):
         return self._pos
@@ -460,9 +497,9 @@ class FileStorage(
 
     def store(self, oid, oldserial, data, version, transaction):
         if self._is_read_only:
-            raise POSException.ReadOnlyError()
+            raise ReadOnlyError()
         if transaction is not self._transaction:
-            raise POSException.StorageTransactionError(self, transaction)
+            raise StorageTransactionError(self, transaction)
         assert not version
 
         with self._lock:
@@ -493,25 +530,25 @@ class FileStorage(
                     "The storage quota has been exceeded.")
 
             if old and oldserial != committed_tid:
-                return ConflictResolution.ResolvedSerial
+                return ResolvedSerial
             else:
                 return self._tid
 
     def deleteObject(self, oid, oldserial, transaction):
         if self._is_read_only:
-            raise POSException.ReadOnlyError()
+            raise ReadOnlyError()
         if transaction is not self._transaction:
-            raise POSException.StorageTransactionError(self, transaction)
+            raise StorageTransactionError(self, transaction)
 
         with self._lock:
             old = self._index_get(oid, 0)
             if not old:
-                raise POSException.POSKeyError(oid)
+                raise POSKeyError(oid)
             h = self._read_data_header(old, oid)
             committed_tid = h.tid
 
             if oldserial != committed_tid:
-                raise POSException.ConflictError(
+                raise ConflictError(
                     oid=oid, serials=(committed_tid, oldserial))
 
             pos = self._pos
@@ -543,6 +580,7 @@ class FileStorage(
         self._file.seek(tpos)
         h = self._file.read(TRANS_HDR_LEN)
         tid, tl, status, ul, dl, el = unpack(TRANS_HDR, h)
+        status = as_text(status)
         self._file.read(ul + dl + el)
         tend = tpos + tl + 8
         pos = self._file.tell()
@@ -591,9 +629,9 @@ class FileStorage(
         # should be considered just a hint, and is ignored if the transaction
         # doesn't exist.
         if self._is_read_only:
-            raise POSException.ReadOnlyError()
+            raise ReadOnlyError()
         if transaction is not self._transaction:
-            raise POSException.StorageTransactionError(self, transaction)
+            raise StorageTransactionError(self, transaction)
         if version:
             raise TypeError("Versions are no-longer supported")
 
@@ -659,7 +697,7 @@ class FileStorage(
     def tpc_vote(self, transaction):
         with self._lock:
             if transaction is not self._transaction:
-                raise POSException.StorageTransactionError(
+                raise StorageTransactionError(
                     "tpc_vote called with wrong transaction")
             dlen = self._tfile.tell()
             if not dlen:
@@ -677,7 +715,7 @@ class FileStorage(
                 h.descr = descr
                 h.ext = ext
                 self._file.write(h.asString())
-                ZODB.utils.cp(self._tfile, self._file, dlen)
+                cp(self._tfile, self._file, dlen)
                 self._file.write(p64(tl))
                 self._file.flush()
             except:
@@ -691,7 +729,7 @@ class FileStorage(
         with self._files.write_lock():
             with self._lock:
                 if transaction is not self._transaction:
-                    raise POSException.StorageTransactionError(
+                    raise StorageTransactionError(
                         "tpc_finish called with wrong transaction")
                 try:
                     if f is not None:
@@ -710,7 +748,7 @@ class FileStorage(
         if self._nextpos:
             # Clear the checkpoint flag
             self._file.seek(self._pos+16)
-            self._file.write(self._tstatus)
+            self._file.write(as_bytes(self._tstatus))
             try:
                 # At this point, we may have committed the data to disk.
                 # If we fail from here, we're in bad shape.
@@ -840,7 +878,7 @@ class FileStorage(
         try:
             data = self.tryToResolveConflict(oid, ctid, tid, bdata, cdata)
             return data, 0, ipos
-        except POSException.ConflictError:
+        except ConflictError:
             pass
 
         raise UndoError("Some data were modified by a later transaction", oid)
@@ -889,13 +927,13 @@ class FileStorage(
         """
 
         if self._is_read_only:
-            raise POSException.ReadOnlyError()
+            raise ReadOnlyError()
         if transaction is not self._transaction:
-            raise POSException.StorageTransactionError(self, transaction)
+            raise StorageTransactionError(self, transaction)
 
         with self._lock:
           # Find the right transaction to undo and call _txn_undo_write().
-          tid = base64.decodestring(transaction_id + '\n')
+          tid = decodebytes(transaction_id + b'\n')
           assert len(tid) == 8
           tpos = self._txn_find(tid, 1)
           tindex = self._txn_undo_write(tpos)
@@ -914,7 +952,7 @@ class FileStorage(
                 return pos
             if stop_at_pack:
                 # check the status field of the transaction header
-                if h[16] == 'p':
+                if h[16] == b'p':
                     break
         raise UndoError("Invalid transaction id")
 
@@ -945,7 +983,7 @@ class FileStorage(
             try:
                 p, prev, ipos = self._transactionalUndoRecord(
                     h.oid, pos, h.tid, h.prev)
-            except UndoError, v:
+            except UndoError as v:
                 # Don't fail right away. We may be redeemed later!
                 failures[h.oid] = v
             else:
@@ -953,16 +991,16 @@ class FileStorage(
                 if self.blob_dir and not p and prev:
                     try:
                         up, userial = self._loadBackTxn(h.oid, prev)
-                    except ZODB.POSException.POSKeyError:
+                    except POSKeyError:
                         pass # It was removed, so no need to copy data
                     else:
                         if self.is_blob_record(up):
                             # We're undoing a blob modification operation.
                             # We have to copy the blob data
-                            tmp = ZODB.utils.mktemp(dir=self.fshelper.temp_dir)
-                            ZODB.utils.cp(
-                                self.openCommittedBlobFile(h.oid, userial),
-                                open(tmp, 'wb'))
+                            tmp = mktemp(dir=self.fshelper.temp_dir)
+                            with self.openCommittedBlobFile(h.oid, userial) as sfp:
+                                with open(tmp, 'wb') as dfp:
+                                    cp(sfp, dfp)
                             self._blob_storeblob(h.oid, self._tid, tmp)
 
                 new = DataHeader(h.oid, self._tid, ipos, otloc, 0, len(p))
@@ -984,7 +1022,7 @@ class FileStorage(
                 raise UndoError("non-undoable transaction")
 
         if failures:
-            raise MultipleUndoErrors(failures.items())
+            raise MultipleUndoErrors(list(failures.items()))
 
         return tindex
 
@@ -1032,10 +1070,13 @@ class FileStorage(
         # want to invest much in the old packer, at least for now.
         assert referencesf is not None
         p = FileStoragePacker(storage, referencesf, stop, gc)
-        opos = p.pack()
-        if opos is None:
-            return None
-        return opos, p.index
+        try:
+            opos = p.pack()
+            if opos is None:
+                return None
+            return opos, p.index
+        finally:
+            p.close()
 
     def pack(self, t, referencesf, gc=None):
         """Copy data from the current database file to a packed file
@@ -1047,10 +1088,11 @@ class FileStorage(
         the associated data are copied, since the old records are not copied.
         """
         if self._is_read_only:
-            raise POSException.ReadOnlyError()
+            raise ReadOnlyError()
 
-        stop=`TimeStamp(*time.gmtime(t)[:5]+(t%60,))`
-        if stop==z64: raise FileStorageError('Invalid pack time')
+        stop = TimeStamp(*time.gmtime(t)[:5]+(t%60,)).raw()
+        if stop == z64:
+            raise FileStorageError('Invalid pack time')
 
         # If the storage is empty, there's nothing to do.
         if not self._index:
@@ -1068,7 +1110,7 @@ class FileStorage(
         if os.path.exists(oldpath):
             os.remove(oldpath)
         if self.blob_dir and os.path.exists(self.blob_dir + ".old"):
-            ZODB.blob.remove_committed_dir(self.blob_dir + ".old")
+            remove_committed_dir(self.blob_dir + ".old")
 
         cleanup = []
 
@@ -1077,7 +1119,7 @@ class FileStorage(
             pack_result = None
             try:
                 pack_result = self.packer(self, referencesf, stop, gc)
-            except RedundantPackWarning, detail:
+            except RedundantPackWarning as detail:
                 logger.info(str(detail))
             if pack_result is None:
                 return
@@ -1123,7 +1165,6 @@ class FileStorage(
         lblob_dir = len(self.blob_dir)
         fshelper = self.fshelper
         old = self.blob_dir+'.old'
-        link_or_copy = ZODB.blob.link_or_copy
 
         # Helper to clean up dirs left empty after moving things to old
         def maybe_remove_empty_dir_containing(path, level=0):
@@ -1161,46 +1202,47 @@ class FileStorage(
 
         if self.pack_keep_old:
             # Helpers that move oid dir or revision file to the old dir.
-            os.mkdir(old, 0777)
+            os.mkdir(old, 0o777)
             link_or_copy(os.path.join(self.blob_dir, '.layout'),
                          os.path.join(old, '.layout'))
             def handle_file(path):
                 newpath = old+path[lblob_dir:]
                 dest = os.path.dirname(newpath)
                 if not os.path.exists(dest):
-                    os.makedirs(dest, 0700)
+                    os.makedirs(dest, 0o700)
                 os.rename(path, newpath)
             handle_dir = handle_file
         else:
             # Helpers that remove an oid dir or revision file.
-            handle_file = ZODB.blob.remove_committed
-            handle_dir = ZODB.blob.remove_committed_dir
+            handle_file = remove_committed
+            handle_dir = remove_committed_dir
 
         # Fist step: move or remove oids or revisions
-        for line in open(os.path.join(self.blob_dir, '.removed')):
-            line = line.strip().decode('hex')
+        with open(os.path.join(self.blob_dir, '.removed'), 'rb') as fp:
+            for line in fp:
+                line = binascii.unhexlify(line.strip())
 
-            if len(line) == 8:
-                # oid is garbage, re/move dir
-                path = fshelper.getPathForOID(line)
+                if len(line) == 8:
+                    # oid is garbage, re/move dir
+                    path = fshelper.getPathForOID(line)
+                    if not os.path.exists(path):
+                        # Hm, already gone. Odd.
+                        continue
+                    handle_dir(path)
+                    maybe_remove_empty_dir_containing(path, 1)
+                    continue
+
+                if len(line) != 16:
+                    raise ValueError("Bad record in ", self.blob_dir, '.removed')
+
+                oid, tid = line[:8], line[8:]
+                path = fshelper.getBlobFilename(oid, tid)
                 if not os.path.exists(path):
                     # Hm, already gone. Odd.
                     continue
-                handle_dir(path)
-                maybe_remove_empty_dir_containing(path, 1)
-                continue
-
-            if len(line) != 16:
-                raise ValueError("Bad record in ", self.blob_dir, '.removed')
-
-            oid, tid = line[:8], line[8:]
-            path = fshelper.getBlobFilename(oid, tid)
-            if not os.path.exists(path):
-                # Hm, already gone. Odd.
-                continue
-            handle_file(path)
-            assert not os.path.exists(path)
-            maybe_remove_empty_dir_containing(path)
+                handle_file(path)
+                assert not os.path.exists(path)
+                maybe_remove_empty_dir_containing(path)
 
         os.remove(os.path.join(self.blob_dir, '.removed'))
 
@@ -1215,7 +1257,7 @@ class FileStorage(
                 file_path = os.path.join(path, file_name)
                 dest = os.path.dirname(old+file_path[lblob_dir:])
                 if not os.path.exists(dest):
-                    os.makedirs(dest, 0700)
+                    os.makedirs(dest, 0o700)
                 link_or_copy(file_path, old+file_path[lblob_dir:])
 
     def iterator(self, start=None, stop=None):
@@ -1252,7 +1294,7 @@ class FileStorage(
         for ext in '', '.old', '.tmp', '.lock', '.index', '.pack':
             try:
                 os.remove(self._file_name + ext)
-            except OSError, e:
+            except OSError as e:
                 if e.errno != errno.ENOENT:
                     raise
 
@@ -1316,6 +1358,7 @@ def shift_transactions_forward(index, tindex, file, pos, opos):
         h=read(TRANS_HDR_LEN)
         if len(h) < TRANS_HDR_LEN: break
         tid, stl, status, ul, dl, el = unpack(TRANS_HDR,h)
+        status = as_text(status)
         if status=='c': break # Oops. we found a checkpoint flag.
         tl=u64(stl)
         tpos=pos
@@ -1403,7 +1446,7 @@ def recover(file_name):
 
     pos, oid, tid = read_index(file, file_name, index, tindex, recover=1)
     if oid is not None:
-        print "Nothing to recover"
+        print("Nothing to recover")
         return
 
     opos=pos
@@ -1413,13 +1456,13 @@ def recover(file_name):
 
     file.truncate(npos)
 
-    print "Recovered file, lost %s, ended up with %s bytes" % (
-        pos-opos, npos)
+    print("Recovered file, lost %s, ended up with %s bytes" % (
+        pos-opos, npos))
 
 
 
-def read_index(file, name, index, tindex, stop='\377'*8,
-               ltid=z64, start=4L, maxoid=z64, recover=0, read_only=0):
+def read_index(file, name, index, tindex, stop=b'\377'*8,
+               ltid=z64, start=4, maxoid=z64, recover=0, read_only=0):
     """Scan the file storage and update the index.
 
     Returns file position, max oid, and last transaction id.  It also
@@ -1466,13 +1509,13 @@ def read_index(file, name, index, tindex, stop='\377'*8,
     else:
         if not read_only:
             file.write(packed_version)
-        return 4L, z64, ltid
+        return 4, z64, ltid
 
     index_get = index.get
 
     pos = start
     seek(start)
-    tid = '\0' * 7 + '\1'
+    tid = b'\0' * 7 + b'\1'
 
     while 1:
         # Read the transaction record
@@ -1487,6 +1530,7 @@ def read_index(file, name, index, tindex, stop='\377'*8,
             break
 
         tid, tl, status, ul, dl, el = unpack(TRANS_HDR, h)
+        status = as_text(status)
 
         if tid <= ltid:
             logger.warning("%s time-stamp reduction at %s", name, pos)
@@ -1617,13 +1661,12 @@ def _truncate(file, name, pos):
                                name, oname)
                 o = open(oname,'wb')
                 file.seek(pos)
-                ZODB.utils.cp(file, o, file_size-pos)
+                cp(file, o, file_size-pos)
                 o.close()
                 break
     except:
-        logger.error("couldn\'t write truncated data for %s", name,
-              exc_info=True)
-        raise POSException.StorageSystemError("Couldn't save truncated data")
+        logger.exception("couldn\'t write truncated data for %s", name)
+        raise StorageSystemError("Couldn't save truncated data")
 
     file.seek(pos)
     file.truncate()
@@ -1635,8 +1678,8 @@ class FileIterator(FileStorageFormatter):
     _ltid = z64
     _file = None
 
-    def __init__(self, filename, start=None, stop=None, pos=4L):
-        assert isinstance(filename, str)
+    def __init__(self, filename, start=None, stop=None, pos=4):
+        assert isinstance(filename, STRING_TYPES)
         file = open(filename, 'rb')
         self._file = file
         self._file_name = filename
@@ -1648,8 +1691,8 @@ class FileIterator(FileStorageFormatter):
             raise ValueError("Given position is greater than the file size",
                              pos, self._file_size)
         self._pos = pos
-        assert start is None or isinstance(start, str)
-        assert stop is None or isinstance(stop, str)
+        assert start is None or isinstance(start, bytes)
+        assert stop is None or isinstance(stop, bytes)
         self._start = start
         self._stop = stop
         if start:
@@ -1665,7 +1708,7 @@ class FileIterator(FileStorageFormatter):
         # implementation.  So just return 0.
         return 0
 
-    # This allows us to pass an iterator as the `other' argument to
+    # This allows us to pass an iterator as the `other` argument to
     # copyTransactionsFrom() in BaseStorage.  The advantage here is that we
     # can create the iterator manually, e.g. setting start and stop, and then
     # just let copyTransactionsFrom() do its thing.
@@ -1682,7 +1725,7 @@ class FileIterator(FileStorageFormatter):
         file = self._file
         pos1 = self._pos
         file.seek(pos1)
-        tid1 = file.read(8)
+        tid1 = file.read(8) # XXX bytes
         if len(tid1) < 8:
             raise CorruptedError("Couldn't read tid.")
         if start < tid1:
@@ -1721,9 +1764,9 @@ class FileIterator(FileStorageFormatter):
                     self._pos = self._file_size
                 return
 
-        t1 = ZODB.TimeStamp.TimeStamp(tid1).timeTime()
-        t2 = ZODB.TimeStamp.TimeStamp(tid2).timeTime()
-        ts = ZODB.TimeStamp.TimeStamp(start).timeTime()
+        t1 = TimeStamp(tid1).timeTime()
+        t2 = TimeStamp(tid2).timeTime()
+        ts = TimeStamp(start).timeTime()
         if (ts - t1) < (t2 - ts):
             return self._scan_forward(pos1, start)
         else:
@@ -1751,7 +1794,7 @@ class FileIterator(FileStorageFormatter):
         while 1:
             pos -= 8
             seek(pos)
-            tlen = ZODB.utils.u64(read(8))
+            tlen = u64(read(8))
             pos -= tlen
             h = self._read_txn_header(pos)
             if h.tid <= start:
@@ -1765,7 +1808,7 @@ class FileIterator(FileStorageFormatter):
     def __iter__(self):
         return self
 
-    def next(self):
+    def __next__(self):
         if self._file is None:
             raise StopIteration()
 
@@ -1775,7 +1818,7 @@ class FileIterator(FileStorageFormatter):
             # Read the transaction record
             try:
                 h = self._read_txn_header(pos)
-            except CorruptedDataError, err:
+            except CorruptedDataError as err:
                 # If buf is empty, we've reached EOF.
                 if not err.buf:
                     break
@@ -1858,11 +1901,13 @@ class FileIterator(FileStorageFormatter):
         self.close()
         raise StopIteration()
 
+    next = __next__
 
-class TransactionRecord(BaseStorage.TransactionRecord):
+
+class TransactionRecord(_TransactionRecord):
 
     def __init__(self, tid, status, user, desc, ext, pos, tend, file, tpos):
-        BaseStorage.TransactionRecord.__init__(
+        _TransactionRecord.__init__(
             self, tid, status, user, desc, ext)
         self._pos = pos
         self._tend = tend
@@ -1884,7 +1929,7 @@ class TransactionRecordIterator(FileStorageFormatter):
     def __iter__(self):
         return self
 
-    def next(self):
+    def __next__(self):
         pos = self._pos
         while pos < self._tend:
             # Read the data records for this transaction
@@ -1917,8 +1962,10 @@ class TransactionRecordIterator(FileStorageFormatter):
 
         raise StopIteration()
 
+    next = __next__
 
-class Record(BaseStorage.DataRecord):
+
+class Record(_DataRecord):
 
     def __init__(self, oid, tid, data, prev, pos):
         super(Record, self).__init__(oid, tid, data, prev)
@@ -1961,12 +2008,13 @@ class UndoSearch:
         self.file.seek(self.pos)
         h = self.file.read(TRANS_HDR_LEN)
         tid, tl, status, ul, dl, el = unpack(TRANS_HDR, h)
+        status = as_text(status)
         if status == 'p':
             self.stop = 1
             return None
         if status != ' ':
             return None
-        d = u = ''
+        d = u = b''
         if ul:
             u = self.file.read(ul)
         if dl:
@@ -1977,7 +2025,7 @@ class UndoSearch:
                 e = loads(self.file.read(el))
             except:
                 pass
-        d = {'id': base64.encodestring(tid).rstrip(),
+        d = {'id': encodebytes(tid).rstrip(),
              'time': TimeStamp(tid).timeTime(),
              'user_name': u,
              'size': tl,
