@@ -13,15 +13,25 @@
 ##############################################################################
 
 from ZODB.tests.MinPO import MinPO
-import datetime
 import doctest
 import os
+import re
 import sys
 import time
 import transaction
 import unittest
 import ZODB
 import ZODB.tests.util
+from zope.testing import renormalizing
+
+checker = renormalizing.RENormalizing([
+    # Python 3 bytes add a "b".
+    (re.compile("b('.*?')"),
+     r"\1"),
+    # Python 3 adds module name to exceptions.
+    (re.compile("ZODB.POSException.ReadConflictError"), r"ReadConflictError"),
+    ])
+
 
 # Return total number of connections across all pools in a db._pools.
 def nconn(pools):
@@ -60,10 +70,10 @@ class DBTests(ZODB.tests.util.TestCase):
     def test_references(self):
 
         # TODO: For now test that we're using referencesf.  We really should
-        #       have tests of referencesf.  
+        #       have tests of referencesf.
 
         import ZODB.serialize
-        self.assert_(self.db.references is ZODB.serialize.referencesf)
+        self.assertTrue(self.db.references is ZODB.serialize.referencesf)
 
 
 def test_invalidateCache():
@@ -99,7 +109,7 @@ def test_invalidateCache():
 
         >>> c3 is db.open(transaction_manager=tm3)
         True
-        >>> print c3.root()['a']._p_changed
+        >>> print(c3.root()['a']._p_changed)
         None
 
         >>> db.close()
@@ -131,17 +141,16 @@ def connectionDebugInfo():
     >>> c3 = db.open(before=c1.root()._p_serial)
 
     >>> info = db.connectionDebugInfo()
-    >>> import pprint
-    >>> pprint.pprint(sorted(info, key=lambda i: str(i['opened'])), width=1)
-    [{'before': None,
-      'info': 'test info (2)',
-      'opened': '2008-12-04T20:40:44Z (1.40s)'},
-     {'before': '\x03zY\xd8\xc0m9\xdd',
-      'info': ' (0)',
-      'opened': '2008-12-04T20:40:45Z (0.30s)'},
-     {'before': None,
-      'info': ' (0)',
-      'opened': None}]
+    >>> info = sorted(info, key=lambda i: str(i['opened']))
+    >>> before = [x['before'] for x in info]
+    >>> opened = [x['opened'] for x in info]
+    >>> infos = [x['info'] for x in info]
+    >>> before
+    [None, '\x03zY\xd8\xc0m9\xdd', None]
+    >>> opened
+    ['2008-12-04T20:40:44Z (1.40s)', '2008-12-04T20:40:45Z (0.30s)', None]
+    >>> infos
+    ['test info (2)', ' (0)', ' (0)']
 
     >>> time.time = real_time
 
@@ -197,13 +206,13 @@ def open_convenience():
     DB arguments.
 
     >>> conn = ZODB.connection('data.fs', blob_dir='blobs')
-    >>> conn.root()['b'] = ZODB.blob.Blob('test')
+    >>> conn.root()['b'] = ZODB.blob.Blob(b'test')
     >>> transaction.commit()
     >>> conn.close()
 
     >>> db = ZODB.DB('data.fs', blob_dir='blobs')
     >>> conn = db.open()
-    >>> conn.root()['b'].open().read()
+    >>> with conn.root()['b'].open() as fp: fp.read()
     'test'
     >>> db.close()
 
@@ -345,33 +354,10 @@ def minimally_test_connection_timeout():
 
     """
 
-def saving_oid_multiple_times_doesnt_cause_dups():
-    r"""Duplicate saves of an oid shouldn't happen unless there's a bug elsewhere
-
-    But saving dups makes matters worse, because it dooms the process,
-    not just the transaction.
-
-    >>> db = ZODB.DB(None)
-    >>> oid = db.new_oid()
-    >>> db.save_oid(oid)
-    >>> db.new_oid() is oid
-    True
-    >>> db.save_oid(oid)
-    >>> db.save_oid(oid)
-    Traceback (most recent call last):
-    ...
-    ValueError: Duplicate saved object ids.
-
-    >>> db.new_oid() is oid
-    True
-    >>> db.new_oid()
-    '\x00\x00\x00\x00\x00\x00\x00\x02'
-
-    """
-
 def test_suite():
     s = unittest.makeSuite(DBTests)
     s.addTest(doctest.DocTestSuite(
         setUp=ZODB.tests.util.setUp, tearDown=ZODB.tests.util.tearDown,
+        checker=checker
         ))
     return s
