@@ -11,45 +11,36 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-import unittest
-import warnings
+from persistent import Persistent
+from persistent.mapping import PersistentMapping
+from ZODB.POSException import ReadConflictError
+from ZODB.POSException import TransactionFailedError
 
+import doctest
+from BTrees.OOBTree import OOBTree
+import transaction
+import unittest
 import ZODB
 import ZODB.FileStorage
 import ZODB.MappingStorage
-from ZODB.POSException import ReadConflictError, ConflictError
-from ZODB.POSException import TransactionFailedError
-from ZODB.tests.warnhook import WarningsHook
-
-from persistent import Persistent
-from persistent.mapping import PersistentMapping
-import transaction
-
-# deprecated39  remove when versions go away
-warnings.filterwarnings("ignore",
-                        "Versions are deprecated",
-                        DeprecationWarning, __name__)
+import ZODB.tests.util
 
 class P(Persistent):
     pass
 
-class Independent(Persistent):
 
-    def _p_independent(self):
-        return 1
-
-class DecoyIndependent(Persistent):
-
-    def _p_independent(self):
-        return 0
-
-class ZODBTests(unittest.TestCase):
+class ZODBTests(ZODB.tests.util.TestCase):
 
     def setUp(self):
+        ZODB.tests.util.TestCase.setUp(self)
         self._storage = ZODB.FileStorage.FileStorage(
             'ZODBTests.fs', create=1)
         self._db = ZODB.DB(self._storage)
 
+    def tearDown(self):
+        self._db.close()
+        ZODB.tests.util.TestCase.tearDown(self)
+
     def populate(self):
         transaction.begin()
         conn = self._db.open()
@@ -61,10 +52,6 @@ class ZODBTests(unittest.TestCase):
         transaction.commit()
         conn.close()
 
-    def tearDown(self):
-        self._db.close()
-        self._storage.cleanup()
-
     def checkExportImport(self, abort_it=False):
         self.populate()
         conn = self._db.open()
@@ -86,14 +73,13 @@ class ZODBTests(unittest.TestCase):
         assert len(ob) > 10, 'Insufficient test data'
         try:
             import tempfile
-            f = tempfile.TemporaryFile()
-            ob._p_jar.exportFile(ob._p_oid, f)
-            assert f.tell() > 0, 'Did not export correctly'
-            f.seek(0)
-            new_ob = ob._p_jar.importFile(f)
-            self.assertEqual(new_ob, ob)
-            root['dup'] = new_ob
-            f.close()
+            with tempfile.TemporaryFile(prefix="DUP") as f:
+                ob._p_jar.exportFile(ob._p_oid, f)
+                assert f.tell() > 0, 'Did not export correctly'
+                f.seek(0)
+                new_ob = ob._p_jar.importFile(f)
+                self.assertEqual(new_ob, ob)
+                root['dup'] = new_ob
             if abort_it:
                 transaction.abort()
             else:
@@ -115,41 +101,27 @@ class ZODBTests(unittest.TestCase):
             else:
                 raise
         else:
-            self.failUnless(not abort_it, 'Did not abort duplication')
+            self.assertTrue(not abort_it, 'Did not abort duplication')
         l1 = list(ob.items())
         l1.sort()
         l2 = list(ob2.items())
         l2.sort()
-        l1 = map(lambda (k, v): (k, v[0]), l1)
-        l2 = map(lambda (k, v): (k, v[0]), l2)
+        l1 = list(map(lambda k_v: (k_v[0], k_v[1][0]), l1))
+        l2 = list(map(lambda k_v1: (k_v1[0], k_v1[1][0]), l2))
         self.assertEqual(l1, l2)
-        self.assert_(ob._p_oid != ob2._p_oid)
+        self.assertTrue(ob._p_oid != ob2._p_oid)
         self.assertEqual(ob._p_jar, ob2._p_jar)
         oids = {}
         for v in ob.values():
             oids[v._p_oid] = 1
         for v in ob2.values():
-            assert not oids.has_key(v._p_oid), (
+            assert v._p_oid not in oids, (
                 'Did not fully separate duplicate from original')
         transaction.commit()
 
     def checkExportImportAborted(self):
         self.checkExportImport(abort_it=True)
 
-    def checkVersionOnly(self):
-        # Make sure the changes to make empty transactions a no-op
-        # still allow things like abortVersion().  This should work
-        # because abortVersion() calls tpc_begin() itself.
-        conn = self._db.open("version")
-        try:
-            r = conn.root()
-            r[1] = 1
-            transaction.commit()
-        finally:
-            conn.close()
-        self._db.abortVersion("version")
-        transaction.commit()
-
     def checkResetCache(self):
         # The cache size after a reset should be 0.  Note that
         # _resetCache is not a public API, but the resetCaches()
@@ -158,7 +130,7 @@ class ZODBTests(unittest.TestCase):
         self.populate()
         conn = self._db.open()
         conn.root()
-        self.assert_(len(conn._cache) > 0)  # Precondition
+        self.assertTrue(len(conn._cache) > 0)  # Precondition
         conn._resetCache()
         self.assertEqual(len(conn._cache), 0)
 
@@ -168,10 +140,10 @@ class ZODBTests(unittest.TestCase):
         self.populate()
         conn = self._db.open()
         conn.root()
-        self.assert_(len(conn._cache) > 0)  # Precondition
+        self.assertTrue(len(conn._cache) > 0)  # Precondition
         ZODB.Connection.resetCaches()
         conn.close()
-        self.assert_(len(conn._cache) > 0)  # Still not flushed
+        self.assertTrue(len(conn._cache) > 0)  # Still not flushed
         conn.open()  # simulate the connection being reopened
         self.assertEqual(len(conn._cache), 0)
 
@@ -185,7 +157,7 @@ class ZODBTests(unittest.TestCase):
         try:
             r1 = conn1.root()
             r2 = conn2.root()
-            if r1.has_key('item'):
+            if 'item' in r1:
                 del r1['item']
                 tm1.get().commit()
             r1.get('item')
@@ -357,6 +329,28 @@ class ZODBTests(unittest.TestCase):
 
         cn.close()
 
+    def checkSavepointRollbackAndReadCurrent(self):
+        '''
+        savepoint rollback after readcurrent was called on a new object
+        should not raise POSKeyError
+        '''
+        cn = self._db.open()
+        try:
+            transaction.begin()
+            root = cn.root()
+            added_before_savepoint = P()
+            root['added_before_savepoint'] = added_before_savepoint
+            sp = transaction.savepoint()
+            added_before_savepoint.btree = new_btree = OOBTree()
+            cn.add(new_btree)
+            new_btree['change_to_trigger_read_current'] = P()
+            sp.rollback()
+            transaction.commit()
+            self.assertTrue('added_before_savepoint' in root)
+        finally:
+            transaction.abort()
+            cn.close()
+
     def checkFailingSavepointSticks(self):
         cn = self._db.open()
         rt = cn.root()
@@ -437,8 +431,8 @@ class ZODBTests(unittest.TestCase):
             # performed yet.
             transaction.begin()
             log = self._db.undoLog()
-            for i in range(5):
-                self._db.undo(log[i]['id'])
+            self._db.undoMultiple([log[i]['id'] for i in range(5)])
+
             transaction.get().note('undo states 1 through 5')
 
             # Now attempt all those undo operations.
@@ -450,9 +444,10 @@ class ZODBTests(unittest.TestCase):
             transaction.abort()
             conn.close()
 
-class ReadConflictTests(unittest.TestCase):
+class ReadConflictTests(ZODB.tests.util.TestCase):
 
     def setUp(self):
+        ZODB.tests.utils.TestCase.setUp(self)
         self._storage = ZODB.MappingStorage.MappingStorage()
 
     def readConflict(self, shouldFail=True):
@@ -510,14 +505,6 @@ class ReadConflictTests(unittest.TestCase):
         self.obj = P()
         self.readConflict()
 
-    def checkIndependent(self):
-        self.obj = Independent()
-        self.readConflict(shouldFail=False)
-
-    def checkNotIndependent(self):
-        self.obj = DecoyIndependent()
-        self.readConflict()
-    
     def checkReadConflictIgnored(self):
         # Test that an application that catches a read conflict and
         # continues can not commit the transaction later.
@@ -553,12 +540,12 @@ class ReadConflictTests(unittest.TestCase):
             self.fail("No conflict occurred")
 
         # real_data2 still ready to commit
-        self.assert_(real_data2._p_changed)
+        self.assertTrue(real_data2._p_changed)
 
         # index2 values not ready to commit
-        self.assert_(not index2._p_changed)
-        self.assert_(not index2[0]._p_changed)
-        self.assert_(not index2[1]._p_changed)
+        self.assertTrue(not index2._p_changed)
+        self.assertTrue(not index2[0]._p_changed)
+        self.assertTrue(not index2[1]._p_changed)
 
         self.assertRaises(ReadConflictError, tm.get().commit)
         self.assertRaises(TransactionFailedError, tm.get().commit)
@@ -596,8 +583,8 @@ class ReadConflictTests(unittest.TestCase):
         # but 3.2.3 had a bug wherein it did.
         data_conflicts = data._p_jar._conflicts
         data2_conflicts = data2._p_jar._conflicts
-        self.failIf(data_conflicts)
-        self.failIf(data2_conflicts)  # this used to fail
+        self.assertFalse(data_conflicts)
+        self.assertFalse(data2_conflicts)  # this used to fail
 
         # And because of that, we still couldn't commit a change to data2['d']
         # in the new transaction.
@@ -647,7 +634,9 @@ class PoisonedObject:
         self._p_jar = poisonedjar
 
 def test_suite():
-    return unittest.makeSuite(ZODBTests, 'check')
+    return unittest.TestSuite((
+        unittest.makeSuite(ZODBTests, 'check'),
+        ))
 
 if __name__ == "__main__":
     unittest.main(defaultTest="test_suite")
